import * as markdownIt from 'markdown-it'
import {insertLineNumbers} from './stringHelper'
import Logger from './logger'
import {copyToClipboard} from './clipboard'

const hljs = require('highlight.js/lib/highlight')

/**
 * all from node_modules/highlight.js/lib/languages
 * @see possibleEditorHighlightPLanguages for editor highlighting
 *
 * some hints about the names can be found here https://highlightjs.org/static/demo/
 */
export const possibleMarkdownHighlightPLanguages: ReadonlyArray<string> = [
  'routeros',
  'ebnf',
  'shell',
  'bash',
  'gauss',
  'delphi',
  'dust',
  'profile',
  'cos',
  'haml',
  'vhdl',
  'mercury',
  'oxygene',
  'apache',
  'rib',
  'prolog',
  'ceylon',
  'powershell',
  'moonscript',
  'perl',
  'typescript',
  'leaf',
  'cmake',
  'cal',
  'smalltalk',
  'cpp',
  'stan',
  'verilog',
  'excel',
  'erb',
  'avrasm',
  'julia-repl',
  'awk',
  'taggerscript',
  'tp',
  'csp',
  'dart',
  'd',
  'pony',
  'matlab',
  'ruby',
  'fsharp',
  'abnf',
  'livescript',
  'xml',
  'livecodeserver',
  'mathematica',
  'sml',
  'gams',
  'crystal',
  'pf',
  'erlang-repl',
  'hsp',
  'capnproto',
  'gherkin',
  'nimrod',
  'htmlbars',
  'hy',
  'scilab',
  'java',
  'objectivec',
  'glsl',
  'parser3',
  'applescript',
  'ada',
  'lua',
  'django',
  'dos',
  'crmsh',
  'lisp',
  'xquery',
  'tex',
  '1c',
  'r',
  'dns',
  'armasm',
  'clean',
  'flix',
  'nix',
  'handlebars',
  'erlang',
  'ruleslanguage',
  'mizar',
  'ocaml',
  'maxima',
  'fortran',
  'elm',
  'haskell',
  'aspectj',
  'bnf',
  'julia',
  'javascript',
  'jboss-cli',
  'golo',
  'q',
  'n1ql',
  'nsis',
  'qml',
  'protobuf',
  'irpf90',
  'subunit',
  'lasso',
  'dts',
  'less',
  'processing',
  'step21',
  'asciidoc',
  'php',
  'ini',
  'sql',
  'vbnet',
  'scala',
  'purebasic',
  'rust',
  'groovy',
  'arduino',
  'dsconfig',
  'gcode',
  'scheme',
  'autoit',
  'x86asm',
  'autohotkey',
  'haxe',
  'clojure',
  'brainfuck',
  'stata',
  'openscad',
  'rsl',
  'json',
  'xl',
  'tcl',
  'tap',
  'stylus',
  'basic',
  'elixir',
  'nginx',
  'twig',
  'zephir',
  'cs',
  'lsl',
  'vbscript',
  'inform7',
  'roboconf',
  'css',
  'smali',
  'mojolicious',
  'yaml',
  'sqf',
  'vala',
  'thrift',
  'llvm',
  'gradle',
  'coffeescript',
  'accesslog',
  'coq',
  'kotlin',
  'diff',
  'axapta',
  'makefile',
  'monkey',
  'python',
  'vim',
  'ldif',
  'dockerfile',
  'go',
  'vbscript-html',
  'mipsasm',
  'http',
  'actionscript',
  'swift',
  'clojure-repl',
  'puppet',
  'markdown',
  'fix',
  'scss',
  'mel',
].sort()


// tslint:disable-next-line
const usedMarkdownHighlightModesCache: {[langKey: string]: boolean} = {}


/**
 * imports the given lang by key async and applies highlighting to all markdonw-body code blocks on the page
 * @param langKey
 */
function importMarkdownHighlightMode(langKey: string): Promise<void> {

  //see https://webpack.js.org/api/module-methods/ for dynamic imports
  if (usedMarkdownHighlightModesCache[langKey]) return Promise.resolve()

  if (possibleMarkdownHighlightPLanguages.indexOf(langKey) === -1) {
    Logger.warn(`unknown markdown highlight lang: ${langKey}`)
    usedMarkdownHighlightModesCache[langKey] = true
    return Promise.resolve()
  }

  const promise = import(
    /* webpackMode: "lazy" */
    /* webpackChunkName: "markdownModes" */
    `highlight.js/lib/languages/${langKey}`).then((langMode) => {

    hljs.registerLanguage(langKey, langMode.default)

    // console.log('loaded mode ' + langKey, langMode)
    //then run hljs again...
    usedMarkdownHighlightModesCache[langKey] = true

    //code block might not be rendered yet so wait a bit... (so probably this is only a problem for large markdown files)
    //even if this fails the next render will have this lang (e.g. after a dialog was shown then the markdown is re rendered)
    setTimeout(() => {
      const allCodeBlocks = document.querySelectorAll('.markdown-body pre code')
      // tslint:disable-next-line
      for (let i = 0; i < allCodeBlocks.length; i++) {
        const codeClock = allCodeBlocks.item(i)
        hljs.highlightBlock(codeClock)
      }

      // tslint:disable-next-line
    }, 500)
  })

  return promise
}


const mdRenderer = markdownIt({
                                html: false,
                                breaks: true,
                                typographer: true,
                                linkify: true,
                                langPrefix: 'lang-',
                                highlight: (code, lang) => {

                                  if (!lang) return code

                                  //only imports if not already done
                                  importMarkdownHighlightMode(lang)

                                  if (!lang || !hljs.getLanguage(lang)) return code
                                  return hljs.highlight(lang, code).value
                                }
                              })
  .use(require('markdown-it-mathjax')())

//https://github.com/markdown-it/markdown-it/blob/master/docs/architecture.md
const oldLinkRule = function (tokens: any, idx: any, options: any, env: any, self: any): any {
  return self.renderToken(tokens, idx, options);
}

mdRenderer.renderer.rules.link_open = (tokens, idx, options, env, self) => {
  tokens[idx].attrPush(['target', '_blank'])
  tokens[idx].attrPush(['rel', 'noopener', 'noreferrer'])
  return oldLinkRule(tokens, idx, options, env, self)
}

const oldCodeFenceRule = mdRenderer.renderer.rules.fence

const spacesAfterLineNumber = '   '

mdRenderer.renderer.rules.fence = (tokens, idx, options, env, self) => {

  const token = tokens[idx]

  const copyText = btoa(token.content)

  // tslint:disable-next-line
  let flags: CodeFenceProperties | null = null

  if (token.info) { //only add line numbers if we specify a code language

    flags = getCodeFenceProperties(token.info)

    if (flags.showLineNumbers) {
      token.content = insertLineNumbers(token.content, spacesAfterLineNumber)
    }

  }

  const preTag = oldCodeFenceRule(tokens, idx, options, env, self)

  //handler is globally defined in index
  const withCopyBtn = `<div class="fenced-code-bock">
  <div class="copy-fenced-code-block">
    <div>${flags
           ? flags.lang
           : ''}</div>
    <div>
      <i class="paste icon clickable" onclick="onCopyFencedCodeBlockClicked(event)" data-copy-text="${copyText}"></i>
    </div>
  </div>
  ${preTag}
</div>`

  return withCopyBtn
}

function getCodeFenceProperties(langLine: string): CodeFenceProperties {

  if (!langLine) return

  //first is the lang

  const prefix = "--"
  const data = langLine.split(" ")
  const codeFenceProperties: CodeFenceProperties = {
    lang: "",
    showLineNumbers: false
  }

  // tslint:disable-next-line
  for (let i = 0; i < data.length; i++) {
    const part = data[i]

    if (i === 0) {
      //this is the lang e.g. java...
      codeFenceProperties.lang = part

    } else if (part === `${prefix}line-numbers`) {
      codeFenceProperties.showLineNumbers = true
    }

  }

  return codeFenceProperties

}

interface CodeFenceProperties {
  // tslint:disable-next-line
  lang: string
  // tslint:disable-next-line
  showLineNumbers: boolean
}

//images are centered with css rule:
/*
.markdown-body img {
  display block
  margin 5px auto
}
 */

export default mdRenderer
